import os
import pytest
import json
import time
import bitjws
from example import server
from flask_bitjws import FlaskBitjws

wif = "KweY4PozGhtkGPMvvD7vk7nLiN6211XZ2QGxLBMginAQW7MBbgp8"


def test_echo_details_request():
    privkey = bitjws.PrivateKey()
    pubkey = bitjws.pubkey_to_addr(privkey.pubkey.serialize())
    echo_msg = {'hello': 'server'}
    data = bitjws.sign_serialize(privkey, data=echo_msg, iat=time.time(), requrl="/echodetails")
    fbj = FlaskBitjws(server.app)
    app = server.app.test_client()
    udata = json.dumps({'username': pubkey[0:8], 'kid': pubkey})
    user = app.post('/user', data=udata)
    echo = app.post('/echodetails', data=data, headers={'Content-Type': 'application/jose'})
    h, p = bitjws.validate_deserialize(echo.get_data().decode('utf8'), requrl='/response')
    assert 'alg' in h
    assert h['alg'] == 'CUSTOM-BITCOIN-SIGN'
    assert 'typ' in h
    assert 'kid' in h
    assert 'headers' in p['data']
    assert 'Content-Type' in p['data']['headers']
    assert p['data']['headers']['Content-Type'] == 'application/jose'
    assert 'jws' in p['data']
    rdata = p['data']['jws']
    assert 'payload' in rdata
    assert rdata['payload']['data'] == echo_msg




"""
The FlaskBitjws source code. Everything fits in this file, for now.
"""
import bitjws
import time
import copy
import urlparse
from flask import Response, current_app
from flask.ext import login


class FlaskUser(login.UserMixin):
    """
    A flask_login UserMixin.
    """

    def __init__(self, username, salt):
        super(FlaskUser, self).__init__()
        self.salt = salt
        self.username = username


def load_user_from_request(req):
    """
    Just like the Flask.login load_user_from_request

    This function performs almost entirely bitjws authentication tasks.
    If you need to customize the user loading from your database,
    the FlaskBitjws.get_user_by_key method is the one to modify.

    :param req: The flask request to load a user based on.
    """
    if "application/jose" in req.headers['content-type']:
        path = urlparse.urlsplit(req.url).path
        for rule in current_app.url_map.iter_rules():
            if path == rule.rule and req.method in rule.methods:
                dedata = req.get_data().decode('utf8')
                req.jws_header, req.jws_payload = \
                    bitjws.validate_deserialize(dedata, requrl=rule.rule)
                break
    if not hasattr(req, 'jws_header') or req.jws_header is None:
        return None

    if (not 'iat' in req.jws_payload or
            req.jws_payload['iat'] <=
            current_app.bitjws.get_last_nonce(req.jws_header['kid'],
                                              req.jws_payload['iat'])):
        return None
    rawu = current_app.bitjws.get_user_by_key(req.jws_header['kid'])
    user = FlaskUser(**rawu)
    return user


class FlaskBitjws(object):
    """
    A wrapper class for the main Flask bitjws application.
    Performs bitjws authentication checks and signs responses.

    Overwrite the methods with real DB calls:

    get_last_nonce
    get_user_by_key
    """

    def __init__(self, app, privkey=None,
                 loginmanager=None):
        """
        Initialize a flask-bitjws Application with optional LoginManager.
        If you do not provide your own LoginManager one will be created.

        Create or set the private key to use with this application.

        :param str app: The flask application
        :param str privkey: the bitjws private key to use for signing responses
        :param flask.ext.login.LoginManager loginmanager: An optional LoginManager
        """
        if privkey is not None and isinstance(privkey, str):
            self._privkey = bitjws.PrivateKey(bitjws.wif_to_privkey(privkey))
        elif privkey is not None and isinstance(privkey, bitjws.PrivateKey):
            self._privkey = privkey
        else:
            self._privkey = bitjws.PrivateKey()
        self.pubkey = bitjws.pubkey_to_addr(self._privkey.pubkey.serialize())
        print "Initializing FlaskBitjws from with address: {}".format(self.pubkey)

        if loginmanager is None:
            loginmanager = login.LoginManager()
            loginmanager.anonymous_user = login.AnonymousUserMixin
            loginmanager.init_app(app)
        loginmanager.request_loader(load_user_from_request)
        app.bitjws = self

    def create_response(self, payload):
        """
        Create a signed bitjws response using the supplied payload.
        The response content-type will be 'application/jose'.

        :param payload: The response content. Must be json-serializable.
        :return: The signed Response with headers.
        :rtype: flask.Response
        """
        signedmess = bitjws.sign_serialize(self._privkey, requrl='/response',
                                           iat=time.time(), data=payload)
        return Response(signedmess, mimetype='application/jose')

    ##############################################################
    # DB STUBS Section
    # Overwrite the remaining methods in this class for your own.
    ##############################################################
    def get_last_nonce(self, key, nonce):
        """
        This method is only an example! Replace it with a real nonce database.

        :param str key: the public key the nonce belongs to
        :param int nonce: the latest nonce
        """
        if not hasattr(self, '_example_nonce_db'):
            # store nonces as a pair {key: lastnonce}
            self._example_nonce_db = {}
        if not key in self._example_nonce_db:
            self._example_nonce_db[key] = nonce
            return 0
        else:
            oldnonce = copy.copy(self._example_nonce_db[key])
            self._example_nonce_db[key] = nonce
            return oldnonce

    def get_user_by_key(self, key):
        """
        This method is only an example! Replace it with a real user database.

        :param str key: the public key the user belongs to
        """
        if not hasattr(self, '_example_user_db'):
            self._example_user_db = {}

        if key in self._example_user_db:
            return self._example_user_db[key]
        return None




import os
import pytest
from flask import Flask
from flask_bitjws import FlaskBitjws
import bitjws
from example import server

wif = "KweY4PozGhtkGPMvvD7vk7nLiN6211XZ2QGxLBMginAQW7MBbgp8"


def test_init_WIF():
    app = Flask(__name__)
    fbj = FlaskBitjws(app, privkey=wif)
    assert wif == bitjws.privkey_to_wif(app.bitjws._privkey.private_key)


def test_init_PrivateKey():
    privkey = bitjws.PrivateKey(bitjws.wif_to_privkey(wif))
    app = Flask(__name__)
    fbj = FlaskBitjws(app, privkey=privkey)
    assert wif == bitjws.privkey_to_wif(app.bitjws._privkey.private_key)




import os
import time
import json
import pytest
import bitjws
from example import server
from flask_bitjws import FlaskBitjws

wif = "KweY4PozGhtkGPMvvD7vk7nLiN6211XZ2QGxLBMginAQW7MBbgp8"


def test_get_bitjws_header_payload():
    app = server.app.test_client()
    coins = app.get('/coin')
    h, p = bitjws.validate_deserialize(coins.get_data().decode('utf8'), requrl='/response')
    assert 'alg' in h
    assert h['alg'] == 'CUSTOM-BITCOIN-SIGN'
    assert 'typ' in h
    assert 'kid' in h
    assert 'data' in p
    assert isinstance(p['data'], list)
    assert 'metal' in p['data'][0] and 'mint' in p['data'][0]


def test_echo_request():
    privkey = bitjws.PrivateKey()
    pubkey = bitjws.pubkey_to_addr(privkey.pubkey.serialize())
    echo_msg = {'hello': 'server'}
    data = bitjws.sign_serialize(privkey, data=echo_msg, iat=time.time(), requrl="/echo")
    fbj = FlaskBitjws(server.app)
    app = server.app.test_client()
    udata = json.dumps({'username': pubkey[0:8], 'kid': pubkey})
    user = app.post('/user', data=udata)
    echo = app.post('/echo', data=data, headers={'Content-Type': 'application/jose'})
    h, p = bitjws.validate_deserialize(echo.get_data().decode('utf8'), requrl='/response')
    assert 'alg' in h
    assert h['alg'] == 'CUSTOM-BITCOIN-SIGN'
    assert 'typ' in h
    assert 'kid' in h
    assert 'data' in p
    assert p['data'] == echo_msg


def test_bad_nonce():
    app = server.app.test_client()
    privkey = bitjws.PrivateKey()
    echo_msg = {'hello': 'server'}
    data = bitjws.sign_serialize(privkey, echo=echo_msg, iat=time.time(), requrl="/echo")
    data2 = bitjws.sign_serialize(privkey, echo=echo_msg, iat=time.time()-2, requrl="/echo")
    echo = app.post('/echo', data=data, headers={'Content-Type': 'application/jose'})
    echo2 = app.post('/echo', data=data2, headers={'Content-Type': 'application/jose'})
    assert echo2.status_code == 401


def test_no_nonce():
    app = server.app.test_client()
    privkey = bitjws.PrivateKey()
    echo_msg = {'hello': 'server'}
    data = bitjws.sign_serialize(privkey, echo=echo_msg, requrl="/echo")
    echo = app.post('/echo', data=data, headers={'Content-Type': 'application/jose'})
    assert echo.status_code == 401


def test_bad_response():
    app = server.app.test_client()
    privkey = bitjws.PrivateKey()
    echo_msg = {'hello': 'server'}
    data = bitjws.sign_serialize(privkey, echo=echo_msg, iat=time.time(), requrl="/echo")
    data2 = bitjws.sign_serialize(privkey, echo='not%s' % echo_msg)
    da = data.split('.')
    da2 = data2.split('.')
    baddata = "%s.%s.%s" % (da[0], da2[1], da[2])
    echo = app.post('/echo', data=data, headers={'Content-Type': 'application/jose'})
    echo.data = baddata
    h, p = bitjws.validate_deserialize(echo.get_data().decode('utf8'), requrl='/response')
    assert h is None


def test_bad_request():
    app = server.app.test_client()
    privkey = bitjws.PrivateKey()
    echo_msg = {'hello': 'server'}
    data = bitjws.sign_serialize(privkey, echo=echo_msg, iat=time.time(), requrl="/echo")
    data2 = bitjws.sign_serialize(privkey, echo='not%s' % echo_msg)
    da = data.split('.')
    da2 = data2.split('.')
    baddata = "%s.%s.%s" % (da[0], da2[1], da[2])
    echo = app.post('/echo', data=baddata, headers={'Content-Type': 'application/jose'})
    assert echo.status_code == 401




from setuptools import setup

classifiers = [
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 2",
    "Topic :: Software Development :: Libraries",
]

setup(
    name='flask-bitjws',
    version='0.1.1',
    py_modules=['flask_bitjws'],
    url='https://github.com/deginner/flask-bitjws',
    license='MIT',
    classifiers=classifiers,
    author='deginner',
    author_email='support@deginner.com',
    description='A Flask extension for bitjws authentication.',
    setup_requires=['pytest-runner'],
    install_requires=[
        "flask>=0.10.0",
        "flask-login",
        'secp256k1==0.11',
        "bitjws==0.6.3.1",
    ],
    tests_require=['pytest', 'pytest-cov']
)

